#
#    DeltaFi - Data transformation and enrichment platform
#
#    Copyright 2021-2023 DeltaFi Contributors <deltafi@deltafi.org>
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

from typing import List

from deltafi.result import LoadResult, ReinjectResult

from .assertions import *
from .compare_helpers import GenericCompareHelper, CompareHelper
from .framework import TestCaseBase, ActionTest, IOContent


class LoadTestCase(TestCaseBase):
    def __init__(self, fields: Dict):
        super().__init__(fields)
        self.metadata = {}
        self.delete_metadata_keys = []
        self.annotations = {}
        self.domains = []
        self.domain_cmp_tool = GenericCompareHelper()
        self.children = []

    def set_domain_compare_tool(self, helper: CompareHelper):
        self.domain_cmp_tool = helper

    def expect_load_result(self, metadata: Dict, delete_metadata_keys: List[str], annotations: Dict, domains: List):
        self.expected_result_type = LoadResult
        self.metadata = metadata
        self.delete_metadata_keys = delete_metadata_keys
        self.annotations = annotations
        self.domains = domains

    def add_reinject_child(self, filename: str, flow: str, content: IOContent, metadata: Dict):
        self.expected_result_type = ReinjectResult
        self.children.append(
            {
                "filename": filename,
                "flow": flow,
                "content": content,
                "metadata": metadata
            }
        )


class LoadActionTest(ActionTest):
    def __init__(self, package_name: str):
        """
        Provides structure for testing DeltaFi Load action
        Args:
            package_name: name of the actions package for finding resources
        """
        super().__init__(package_name)

    def load(self, test_case: LoadTestCase):
        if test_case.expected_result_type == ReinjectResult:
            self.expect_reinject_result(test_case)
        elif test_case.expected_result_type == LoadResult:
            self.expect_load_result(test_case)
        else:
            super().execute(test_case)

    def expect_load_result(self, test_case: LoadTestCase):
        result = super().run_and_check_result_type(test_case, LoadResult)
        self.assert_load_result(test_case, result)

    def expect_reinject_result(self, test_case: LoadTestCase):
        result = super().run_and_check_result_type(test_case, ReinjectResult)
        self.assert_reinject_result(test_case, result)

    def assert_load_result(self, test_case: LoadTestCase, result: LoadResult):
        # Check output
        self.compare_all_output(test_case.compare_tool, result.content)

        # Check metadata
        assert_keys_and_values(test_case.metadata, result.metadata)

        # Check deleted metadata
        for key in test_case.delete_metadata_keys:
            assert_key_in(key, result.delete_metadata_keys)

        # Check annotations
        assert_keys_and_values(test_case.annotations, result.annotations)

        # Check domains
        self.compare_domains(test_case.domain_cmp_tool, test_case.domains, result.domains)

    def assert_reinject_result(self, test_case: LoadTestCase, actual: ReinjectResult):
        assert_equal_len(test_case.children, actual.children)
        for index, expected in enumerate(test_case.children):
            reinject_child = actual.children[index]
            assert_equal(expected['filename'], reinject_child.filename)
            assert_equal(expected['flow'], reinject_child.flow)
            assert_keys_and_values(expected['metadata'], reinject_child.metadata)

            expected_value = expected['content']
            child_content = reinject_child.content[0]
            seg_id = child_content.segments[0].uuid
            actual_content = self.content_service.get_output(seg_id)

            if type(expected_value) == str:
                test_case.domain_cmp_tool.compare(expected_value, actual_content, f"RI_child[{index}]")
            elif type(expected_value) == IOContent:
                expected_data = self.load_file(expected_value)
                test_case.domain_cmp_tool.compare(expected_data, actual_content, f"RI_child[{index}]")
            else:
                raise ValueError(
                    f"unknown expected_value type: {type(expected_value)}")



#
#    DeltaFi - Data transformation and enrichment platform
#
#    Copyright 2021-2023 DeltaFi Contributors <deltafi@deltafi.org>
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

from typing import Dict

from .constants import IGNORE_VALUE


def assert_equal(e, a):
    assert e == a, f"{e} != {a}"


def assert_equal_with_label(e, a, l):
    assert e == a, f"{l}. Expected:\n<<{e}>>\nBut was:\n<<{a}>>"


def assert_equal_len(e, a):
    assert len(e) == len(a), f"{len(e)} != {len(a)}"


def assert_key_in(k, m):
    assert k in m, f"{k} not found"


def assert_key_not_in(k, m):
    assert k not in m, f"{k} found, but not expected"


def assert_keys_and_values(expected: Dict, actual: Dict):
    for key in expected:
        assert_key_in(key, actual)
        if expected[key] != IGNORE_VALUE:
            assert_equal(expected[key], actual[key])



#
#    DeltaFi - Data transformation and enrichment platform
#
#    Copyright 2021-2023 DeltaFi Contributors <deltafi@deltafi.org>
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

import json
from abc import ABC
from abc import abstractmethod
from typing import List

from deepdiff import DeepDiff

from .assertions import *


class CompareHelper(ABC):
    @abstractmethod
    def compare(self, expected: str, actual: str, label: str):
        pass


class GenericCompareHelper(CompareHelper):
    def compare(self, expected: str, actual: str, label: str):
        assert_equal_with_label(expected, actual, label)


class JsonCompareHelper(CompareHelper):
    def __init__(self, regex_exclusion_list: List):
        self.excludes = regex_exclusion_list

    def compare(self, expected: str, actual: str, label: str):
        exp = json.loads(expected)
        act = json.loads(actual)
        diffs = DeepDiff(exp, act, exclude_regex_paths=self.excludes)
        if len(diffs) > 0:
            raise ValueError(f"{diffs}")
        assert len(diffs) == 0



#
#    DeltaFi - Data transformation and enrichment platform
#
#    Copyright 2021-2023 DeltaFi Contributors <deltafi@deltafi.org>
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

import uuid
from abc import ABC
from importlib.resources import files
from typing import List

from deltafi.domain import DeltaFileMessage, Event, Content, Context
from deltafi.logger import get_logger
from deltafi.result import ErrorResult, FilterResult
from deltafi.storage import Segment

from .assertions import *
from .compare_helpers import GenericCompareHelper, CompareHelper
from .constants import *


class IOContent:
    """
    The IOContent class holds the details for loading input or output
     content into the test framework.
    Attributes:
        file_name (str): The name of file in test/data.
        content_name (str): The name of the content.
        content_type (str): The media type of the content
        offset (int): Offset to use in Segment
        content_bytes (str): Bypass file read, and uses these bytes for content
    """

    def __init__(self, file_name: str, content_name: str = None,
                 content_type: str = None, offset: int = 0,
                 content_bytes: str = ""):
        self.file_name = file_name
        if content_name is None:
            self.content_name = file_name
        else:
            self.content_name = content_name
        if content_type is None:
            self.content_type = IOContent.file_type(file_name)
        else:
            self.content_type = content_type
        self.offset = offset
        self.content_bytes = content_bytes

    @classmethod
    def file_type(cls, name: str):
        if name.endswith(".json"):
            return "application/json"
        elif name.endswith(".xml"):
            return "application/xnl"
        elif name.endswith(".txt"):
            return "text/plain"
        else:
            return "application/octet-stream"


class LoadedContent:
    def __init__(self, did: str, ioc: IOContent, data: str):
        self.name = ioc.content_name
        self.content_type = ioc.content_type
        self.offset = ioc.offset
        if data is not None:
            self.data = data
        else:
            self.data = ioc.content_bytes
        self.segment = Segment.from_dict({
            "uuid": str(uuid.uuid4()),
            "offset": self.offset,
            "size": len(self.data),
            "did": did
        })


class InternalContentService:
    def __init__(self):
        self.loaded_content = {}
        self.outputs = {}

    def load(self, content_list: List[LoadedContent]):
        for c in content_list:
            self.loaded_content[c.segment.uuid] = c

    def put_str(self, did: str, string_data: str):
        segment = Segment(uuid=str(uuid.uuid4()),
                          offset=0,
                          size=len(string_data),
                          did=did)
        self.outputs[segment.uuid] = string_data
        return segment

    def get_str(self, segments: List[Segment]):
        # TODO: String multiple segment ids together
        seg_id = segments[0].uuid
        return self.loaded_content[seg_id].data

    def get_output(self, seg_id: str):
        return self.outputs[seg_id]


class TestCaseBase(ABC):
    def __init__(self, data: Dict):
        """
        A test case for DeltaFi python actions
        :param data: Dict of test case fields
        - action: instance of the action being tested
        - data_dir: str: subdirectory name (e.g., test name) for locating test data files, i.e., test/data/{data_dir)
        - compare_tool: (optional) CompareHelper instanced for comparing output content
        - inputs: (optional) List[IOContent]: input content to action
        - parameters: (optional) Dict: map of action input parameters
        - in_meta: (optional) Dict: map of metadata as input to action
        - in_domains: (optional) List[Domain]: list of domains as input to action
        - in_enrichments: (optional) List[Domain]: list of enrichments as input to action
        """
        if "action" in data:
            self.action = data["action"]
        else:
            raise ValueError("action is required")

        if "data_dir" in data:
            self.data_dir = data["data_dir"]
        else:
            raise ValueError("data_dir is required")

        if "compare_tool" in data:
            self.compare_tool = data["compare_tool"]
        else:
            self.compare_tool = GenericCompareHelper()

        self.inputs = data["inputs"] if "inputs" in data else []
        self.file_name = data["file_name"] if "file_name" in data else "filename"
        self.outputs = data["outputs"] if "outputs" in data else []
        self.parameters = data["parameters"] if "parameters" in data else {}
        self.in_meta = data["in_meta"] if "in_meta" in data else {}
        self.in_domains = data["in_domains"] if "in_domains" in data else []
        self.in_enrichments = data["in_enrichments"] if "in_enrichments" in data else []
        self.expected_result_type = None
        self.cause_regex = None
        self.context_regex = None

    def expect_error_result(self, cause: str, context: str):
        """
        A Sets the expected output of the action to an Error Result
        :param cause: the expected error cause
        :param context: the expected error context
        """
        self.expected_result_type = ErrorResult
        self.cause_regex = cause
        self.context_regex = context

    def expect_filter_result(self, cause: str, context: str=None):
        """
        A Sets the expected output of the action to a Filter Result
        :param cause: the expected filter cause (message)
        :param context: the expected error context (optional)
        """
        self.expected_result_type = FilterResult
        self.cause_regex = cause
        self.context_regex = context


class ActionTest(ABC):
    def __init__(self, package_name: str):
        """
        Provides structure for testing DeltaFi actions
        Args:
            package_name: name of the actions package for finding resources
        """
        self.content_service = InternalContentService()
        self.did = ""
        self.expected_outputs = []
        self.loaded_inputs = []
        self.package_name = package_name
        self.res_path = ""

    def __reset__(self):
        self.content_service = InternalContentService()
        self.did = str(uuid.uuid4())
        self.expected_outputs = []
        self.loaded_inputs = []
        self.res_path = ""

    def load_file(self, ioc: IOContent):
        file_res = self.res_path.joinpath(ioc.file_name)
        with file_res.open("r") as f:
            contents = f.read()
        return contents

    def get_contents(self, test_case: TestCaseBase):
        pkg_path = files(self.package_name)
        self.res_path = pkg_path.joinpath(f"test/data/{test_case.data_dir}/")

        # Load inputs
        for input_ioc in test_case.inputs:
            if len(input_ioc.content_bytes) == 0:
                self.loaded_inputs.append(LoadedContent(self.did, input_ioc, self.load_file(input_ioc)))
            else:
                self.loaded_inputs.append(LoadedContent(self.did, input_ioc, None))

        # Load expected outputs
        for output_ioc in test_case.outputs:
            if len(output_ioc.content_bytes) == 0:
                self.expected_outputs.append(LoadedContent(self.did, output_ioc, self.load_file(output_ioc)))
            else:
                self.expected_outputs.append(LoadedContent(self.did, output_ioc, None))

    def make_content_list(self, test_case: TestCaseBase):
        content_list = []
        for loaded_input in self.loaded_inputs:
            c = Content(name=loaded_input.name,
                        segments=[loaded_input.segment],
                        media_type=loaded_input.content_type,
                        content_service=self.content_service)
            content_list.append(c)
            loaded_input.content = c

        return content_list

    def make_df_msg(self, test_case: TestCaseBase):
        content_list = self.make_content_list(test_case)
        self.content_service.load(self.loaded_inputs)

        return DeltaFileMessage(metadata=test_case.in_meta,
                                content_list=content_list,
                                domains=test_case.in_domains,
                                enrichments=test_case.in_enrichments)

    def make_context(self, test_case: TestCaseBase):
        action_name = INGRESS_FLOW + "." + test_case.action.__class__.__name__
        return Context(did=self.did,
                       action_flow=INGRESS_FLOW,
                       action_name=action_name,
                       source_filename=test_case.file_name,
                       ingress_flow=INGRESS_FLOW,
                       egress_flow=EGRESS_FLOW,
                       system=SYSTEM,
                       hostname=HOSTNAME,
                       content_service=self.content_service,
                       collect=None,
                       collected_dids=None,
                       logger=get_logger())

    def make_event(self, test_case: TestCaseBase):
        return Event(
            delta_file_messages=[self.make_df_msg(test_case)],
            context=self.make_context(test_case),
            params=test_case.parameters,
            queue_name="",
            return_address="")

    def call_action(self, test_case: TestCaseBase):
        self.get_contents(test_case)
        return test_case.action.execute_action(self.make_event(test_case))

    def run_and_check_result_type(self, test_case: TestCaseBase, result_type):
        self.__reset__()
        result = self.call_action(test_case)

        if not isinstance(result, result_type):
            raise ValueError(f"Result type {result.__class__.__name__} does not match {result_type.__name__}")

        return result

    def execute_error(self, test_case: TestCaseBase):
        result = self.run_and_check_result_type(test_case, ErrorResult)
        resp = result.response()
        assert resp['cause'] == test_case.cause_regex
        assert resp['context'] == test_case.context_regex

    def execute_filter(self, test_case):
        result = self.run_and_check_result_type(test_case, FilterResult)
        resp = result.response()
        assert resp['message'] == test_case.cause_regex

    def execute(self, test_case: TestCaseBase):
        if isinstance(test_case.expected_result_type, ErrorResult.__class__):
            self.execute_error(test_case)
        elif isinstance(test_case.expected_result_type, FilterResult.__class__):
            self.execute_filter(test_case)
        else:
            raise ValueError(f"unknown type: {test_case.expected_result_type}")

    def compare_content_details(self, expected: LoadedContent, actual: Content):
        assert_equal(expected.content_type, actual.media_type)
        assert_equal(expected.name, actual.name)

    def compare_one_content(self, comparitor: CompareHelper, expected: LoadedContent, actual, index):
        self.compare_content_details(expected, actual)
        seg_id = actual.segments[0].uuid
        comparitor.compare(
            expected.data,
            self.content_service.get_output(seg_id),
            f"Content[{index}]"
        )

    def compare_all_output(self, comparitor: CompareHelper, content: List):
        assert_equal_len(self.expected_outputs, content)
        for index, expected in enumerate(self.expected_outputs):
            self.compare_one_content(comparitor, expected, content[index], index)

    def compare_domains(self, comparitor: CompareHelper, expected_items: List[Dict], results: List[Dict]):
        assert_equal_len(expected_items, results)
        for index, expected in enumerate(expected_items):
            actual = results[index]
            assert_equal(expected['name'], actual['name'])
            assert_equal(expected['mediaType'], actual['mediaType'])

            expected_value = expected['value']
            if type(expected_value) == str:
                comparitor.compare(expected_value, actual['value'], f"Domain[{index}]")
            elif type(expected_value) == IOContent:
                expected_data = self.load_file(expected_value)
                comparitor.compare(expected_data, actual['value'], f"Domain[{index}]")
            else:
                raise ValueError(
                    f"unknown expected_value type: {type(expected_value)}")



#
#    DeltaFi - Data transformation and enrichment platform
#
#    Copyright 2021-2023 DeltaFi Contributors <deltafi@deltafi.org>
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

from deltafi.domain import *
from deltafi.exception import MissingMetadataException, ExpectedContentException, MissingDomainException, \
    MissingEnrichmentException


class DomainInput(NamedTuple):
    content: List[Content]
    metadata: Dict[str, str]
    domains: Dict[str, Domain]

    def has_content(self) -> bool:
        return len(self.content) > 0

    def content_at(self, index: int) -> Content:
        if len(self.content) < index + 1:
            raise ExpectedContentException(index, len(self.content))
        return self.content[index]

    def first_content(self):
        return self.content_at(0)

    def get_metadata(self, key: str):
        if key in self.metadata:
            return self.metadata[key]
        else:
            raise MissingMetadataException(key)

    def get_metadata_or_else(self, key: str, default: str) -> str:
        if key in self.metadata:
            return self.metadata[key]
        else:
            return default

    def has_domain(self, name: str) -> bool:
        return name in self.domains

    def domain(self, name: str) -> Domain:
        if not self.has_domain(name):
            raise MissingDomainException(name)
        return self.domains[name]


class EgressInput(NamedTuple):
    content: Content
    metadata: dict


class EnrichInput(NamedTuple):
    content: List[Content]
    metadata: dict
    domains: Dict[str, Domain]
    enrichments: Dict[str, Domain]

    def has_content(self) -> bool:
        return len(self.content) > 0

    def content_at(self, index: int) -> Content:
        if len(self.content) < index + 1:
            raise ExpectedContentException(index, len(self.content))
        return self.content[index]

    def first_content(self):
        return self.content_at(0)

    def get_metadata(self, key: str):
        if key in self.metadata:
            return self.metadata[key]
        else:
            raise MissingMetadataException(key)

    def get_metadata_or_else(self, key: str, default: str) -> str:
        if key in self.metadata:
            return self.metadata[key]
        else:
            return default

    def has_domain(self, name: str) -> bool:
        return name in self.domains

    def domain(self, name: str) -> Domain:
        if not self.has_domain(name):
            raise MissingDomainException(name)
        return self.domains[name]

    def has_enrichment(self, name: str) -> bool:
        return name in self.enrichments

    def enrichment(self, name: str) -> Domain:
        if not self.has_enrichment(name):
            raise MissingEnrichmentException(name)
        return self.enrichments[name]


class FormatInput(NamedTuple):
    content: List[Content]
    metadata: dict
    domains: Dict[str, Domain]
    enrichments: Dict[str, Domain]

    def has_content(self) -> bool:
        return len(self.content) > 0

    def content_at(self, index: int) -> Content:
        if len(self.content) < index + 1:
            raise ExpectedContentException(index, len(self.content))
        return self.content[index]

    def first_content(self):
        return self.content_at(0)

    def get_metadata(self, key: str):
        if key in self.metadata:
            return self.metadata[key]
        else:
            raise MissingMetadataException(key)

    def get_metadata_or_else(self, key: str, default: str) -> str:
        if key in self.metadata:
            return self.metadata[key]
        else:
            return default

    def has_domain(self, name: str) -> bool:
        return name in self.domains

    def domain(self, name: str) -> Domain:
        if not self.has_domain(name):
            raise MissingDomainException(name)
        return self.domains[name]

    def has_enrichment(self, name: str) -> bool:
        return name in self.enrichments

    def enrichment(self, name: str) -> Domain:
        if not self.has_enrichment(name):
            raise MissingEnrichmentException(name)
        return self.enrichments[name]


class LoadInput(NamedTuple):
    content: List[Content]
    metadata: dict

    def has_content(self) -> bool:
        return len(self.content) > 0

    def content_at(self, index: int) -> Content:
        if len(self.content) < index + 1:
            raise ExpectedContentException(index, len(self.content))
        return self.content[index]

    def first_content(self):
        return self.content_at(0)

    def get_metadata(self, key: str):
        if key in self.metadata:
            return self.metadata[key]
        else:
            raise MissingMetadataException(key)

    def get_metadata_or_else(self, key: str, default: str) -> str:
        if key in self.metadata:
            return self.metadata[key]
        else:
            return default


class TransformInput(NamedTuple):
    content: List[Content]
    metadata: dict

    def has_content(self) -> bool:
        return len(self.content) > 0

    def content_at(self, index: int) -> Content:
        if len(self.content) < index + 1:
            raise ExpectedContentException(index, len(self.content))
        return self.content[index]

    def first_content(self):
        return self.content_at(0)

    def get_metadata(self, key: str):
        if key in self.metadata:
            return self.metadata[key]
        else:
            raise MissingMetadataException(key)

    def get_metadata_or_else(self, key: str, default: str) -> str:
        if key in self.metadata:
            return self.metadata[key]
        else:
            return default


class ValidateInput(NamedTuple):
    content: Content
    metadata: dict



#
#    DeltaFi - Data transformation and enrichment platform
#
#    Copyright 2021-2023 DeltaFi Contributors <deltafi@deltafi.org>
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

import json
import os
import sys
import threading
import time
import traceback
from datetime import datetime, timezone, timedelta
from os.path import isdir, isfile, join
from pathlib import Path
from typing import List
import importlib
import inspect
import pkgutil

from importlib import metadata
import requests
from deltafi.actioneventqueue import ActionEventQueue
from deltafi.domain import Event, ActionExecution
from deltafi.exception import ExpectedContentException, MissingDomainException, MissingEnrichmentException, \
    MissingMetadataException
from deltafi.logger import get_logger
from deltafi.result import ErrorResult
from deltafi.storage import ContentService
from deltafi.action import Action


def _coordinates():
    return PluginCoordinates(os.getenv('PROJECT_GROUP'), os.getenv('PROJECT_NAME'), os.getenv('PROJECT_VERSION'))


def _setup_queue(max_connections):
    redis_url = os.getenv('REDIS_URL', 'http://deltafi-redis-master:6379')
    password = os.getenv('REDIS_PASSWORD')
    return ActionEventQueue(redis_url, max_connections, password)


def _setup_content_service():
    minio_url = os.getenv('MINIO_URL', 'http://deltafi-minio:9000')
    return ContentService(minio_url,
                          os.getenv('MINIO_ACCESSKEY'),
                          os.getenv('MINIO_SECRETKEY'))


class PluginCoordinates(object):
    def __init__(self, group_id: str, artifact_id: str, version: str):
        self.group_id = group_id
        self.artifact_id = artifact_id
        self.version = version

    def __json__(self):
        return {
            "groupId": self.group_id,
            "artifactId": self.artifact_id,
            "version": self.version
        }


LONG_RUNNING_TASK_DURATION = timedelta(seconds=5)


class Plugin(object):
    def __init__(self, description: str, plugin_name: str = None, plugin_coordinates: PluginCoordinates = None,
                 actions: List = None, action_package: str = None):
        """
        Initialize the plugin object
        :param plugin_name: Name of the plugin project
        :param description: Description of the plugin
        :param plugin_coordinates: plugin coordinates of the plugin, if None the coordinates must be defined in
                                   environment variables
        :param actions: list of action classes to run
        :param action_package: name of the package containing the actions to run
        """
        self.logger = get_logger()

        self.content_service = None
        self.queue = None
        self.actions = []
        self.core_url = os.getenv('CORE_URL')
        action_classes = []
        if actions is not None and len(actions):
            action_classes.extend(actions)

        if action_package is not None:
            found_actions = Plugin.find_actions(action_package)
            if len(found_actions):
                action_classes.extend(found_actions)

        unique_actions = dict.fromkeys(action_classes)
        self.actions = [action() for action in unique_actions]

        self.description = description
        self.display_name = os.getenv('PROJECT_NAME') if plugin_name is None else plugin_name
        self.coordinates = _coordinates() if plugin_coordinates is None else plugin_coordinates

        if os.getenv('ACTIONS_HOSTNAME'):
            self.hostname = os.getenv('ACTIONS_HOSTNAME')
        elif os.getenv('HOSTNAME'):
            self.hostname = os.getenv('HOSTNAME')
        elif os.getenv('COMPUTERNAME'):
            self.hostname = os.getenv('COMPUTERNAME')
        else:
            self.hostname = 'UNKNOWN'

        self.logger.debug(f"Initialized ActionRunner with actions {self.actions}")

    @staticmethod
    def find_actions(package_name) -> List[object]:
        """
        Find all concrete classes that extend the base Action class in the given package
        :param package_name: name of the package to load and scan for actions
        :return: list of classes that extend the Action class
        """
        package = importlib.import_module(package_name)
        classes = []
        visited = set()

        # Iterate over all submodules in the package
        for _, module_name, _ in pkgutil.walk_packages(package.__path__):
            try:
                module = importlib.import_module(package.__name__ + '.' + module_name)
            except ModuleNotFoundError:
                continue

            # Iterate over all members in the module
            for name, obj in inspect.getmembers(module):
                if inspect.isclass(obj) and obj.__module__.startswith(package_name) and obj not in visited:
                    if Plugin.is_action(obj):
                        classes.append(obj)
                    visited.add(obj)

        return classes

    @staticmethod
    def is_action(maybe_action: type) -> bool:
        """
        Check if the given object is a non-abstract subclass of the Action class
        :param maybe_action: object to inspect to see if it is an Action class
        :return: true if the object is a non-abstract subclass of the Action class
        """
        return not inspect.isabstract(maybe_action) and issubclass(maybe_action, Action)

    def action_name(self, action):
        return f"{self.coordinates.group_id}.{action.__class__.__name__}"

    def _action_json(self, action):
        return {
            'name': self.action_name(action),
            'description': action.description,
            'type': action.action_type.name,
            'requiresDomains': action.requires_domains,
            'requiresEnrichments': action.requires_enrichments,
            'schema': action.param_class().model_json_schema()
        }

    def registration_json(self):
        flows_path = str(Path(os.path.dirname(os.path.abspath(sys.argv[0]))) / 'flows')

        flow_files = []
        variables = []
        if isdir(flows_path):
            flow_files = [f for f in os.listdir(flows_path) if isfile(join(flows_path, f))]
            if 'variables.json' in flow_files:
                flow_files.remove('variables.json')
                variables = json.load(open(join(flows_path, 'variables.json')))
        else:
            self.logger.warning(f"Flows directory ({flows_path}) does not exist. No flows will be installed.")

        flows = [json.load(open(join(flows_path, f))) for f in flow_files]
        actions = [self._action_json(action) for action in self.actions]

        return {
                'pluginCoordinates': self.coordinates.__json__(),
                'displayName': self.display_name,
                'description': self.description,
                'actionKitVersion': metadata.version('deltafi'),
                'dependencies': [],
                'actions': actions,
                'variables': variables,
                'flowPlans': flows
            }

    def _register(self):
        url = f"{self.core_url}/plugins"
        headers = {'Content-type': 'application/json'}
        registration_json = self.registration_json()

        self.logger.info(f"Registering plugin:\n{registration_json}")

        response = requests.post(url, headers=headers, json=registration_json)
        if not response.ok:
            self.logger.error(f"Failed to register plugin ({response.status_code}):\n{response.content}")
            exit(1)

        self.logger.info("Plugin registered")

    def run(self):
        self.logger.info("Plugin starting")
        self.queue = _setup_queue(len(self.actions) + 1)
        self.content_service = _setup_content_service()
        self._register()
        for action in self.actions:
            threading.Thread(target=self._do_action, args=(action,)).start()

        threading.Thread(target=self._heartbeat).start()

        self.logger.info("All threads running")

        f = open("/tmp/running", "w")
        f.close()

        self.logger.info("Application initialization complete")

    def _heartbeat(self):
        long_running_actions = set()
        while True:
            try:
                # Set heartbeats
                for action in self.actions:
                    self.queue.heartbeat(self.action_name(action))

                # Record long running tasks
                new_long_running_actions = set()
                for action in self.actions:
                    if action.action_execution and action.action_execution.exceeds_duration(LONG_RUNNING_TASK_DURATION):
                        action_execution = action.action_execution
                        new_long_running_actions.add(action_execution)
                        self.queue.record_long_running_task(action_execution)

                # Remove old long running tasks
                tasks_to_remove = long_running_actions - new_long_running_actions
                for action_execution in tasks_to_remove:
                    self.queue.remove_long_running_task(action_execution)

                long_running_actions = new_long_running_actions

            except Exception as e:
                self.logger.error(f"Failed to register action queue heartbeat or record long running tasks: {e}", e)
            finally:
                time.sleep(10)

    def _do_action(self, action):
        action_logger = get_logger(self.action_name(action))

        action_logger.info(f"Listening on {self.action_name(action)}")
        while True:
            try:
                event_string = self.queue.take(self.action_name(action))
                event = Event.create(json.loads(event_string), self.hostname, self.content_service, action_logger)
                start_time = time.time()
                action_logger.debug(f"Processing event for did {event.context.did}")

                action.action_execution = ActionExecution(self.action_name(action), event.context.action_name,
                                                          event.context.did, datetime.now(timezone.utc))

                try:
                    result = action.execute_action(event)
                except ExpectedContentException as e:
                    result = ErrorResult(event.context,
                                         f"Action attempted to look up element {e.index + 1} (index {e.index}) from "
                                         f"content list of size {e.size}",
                                         f"{str(e)}\n{traceback.format_exc()}")
                except MissingDomainException as e:
                    result = ErrorResult(event.context,
                                         f"Action attempted to access domain {e.name}, which does not exist",
                                         f"{str(e)}\n{traceback.format_exc()}")
                except MissingEnrichmentException as e:
                    result = ErrorResult(event.context,
                                         f"Action attempted to access enrichment {e.name}, which does not exist",
                                         f"{str(e)}\n{traceback.format_exc()}")
                except MissingMetadataException as e:
                    result = ErrorResult(event.context,
                                         f"Missing metadata with key {e.key}",
                                         f"{str(e)}\n{traceback.format_exc()}")
                except BaseException as e:
                    result = ErrorResult(event.context,
                                         f"Action execution {type(e)} exception", f"{str(e)}\n{traceback.format_exc()}")

                action.action_execution = None

                response = {
                    'did': event.context.did,
                    'action': event.context.action_flow + "." + event.context.action_name,
                    'start': start_time,
                    'stop': time.time(),
                    'type': result.result_type,
                    'metrics': [metric.json() for metric in result.metrics]
                }
                if result.result_key is not None:
                    response[result.result_key] = result.response()

                topic = 'dgs'
                if event.return_address:
                    topic += f"-{event.return_address}"
                self.queue.put(topic, json.dumps(response))
            except BaseException as e:
                action_logger.error(f"Unexpected {type(e)} error: {str(e)}\n{traceback.format_exc()}")
                time.sleep(1)



#
#    DeltaFi - Data transformation and enrichment platform
#
#    Copyright 2021-2023 DeltaFi Contributors <deltafi@deltafi.org>
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

import copy
from datetime import datetime, timedelta, timezone
from logging import Logger
from typing import Dict, List, NamedTuple

from deltafi.storage import ContentService, Segment


class ActionExecution(NamedTuple):
    clazz: str
    action: str
    did: str
    start_time: datetime

    def exceeds_duration(self, duration: timedelta) -> bool:
        return self.start_time + duration < datetime.now(timezone.utc)

    @property
    def key(self) -> str:
        return f"{self.clazz}:{self.action}:{self.did}"


class Context(NamedTuple):
    did: str
    action_flow: str
    action_name: str
    source_filename: str
    ingress_flow: str
    egress_flow: str
    system: str
    hostname: str
    content_service: ContentService
    collect: dict = None
    collected_dids: List[str] = None
    memo: str = None
    logger: Logger = None

    @classmethod
    def create(cls, context: dict, hostname: str, content_service: ContentService, logger: Logger):
        did = context['did']
        action_name_parts = context['name'].split(".")
        action_flow = action_name_parts[0]
        action_name = action_name_parts[1]
        if 'sourceFilename' in context:
            source_filename = context['sourceFilename']
        else:
            source_filename = None
        ingress_flow = context['ingressFlow']
        if 'egressFlow' in context:
            egress_flow = context['egressFlow']
        else:
            egress_flow = None
        system = context['systemName']
        if 'collect' in context:
            collect = context['collect']
        else:
            collect = None
        if 'collectedDids' in context:
            collected_dids = context['collectedDids']
        else:
            collected_dids = None
        if 'memo' in context:
            memo = context['memo']
        else:
            memo = None
        return Context(did=did,
                       action_flow=action_flow,
                       action_name=action_name,
                       source_filename=source_filename,
                       ingress_flow=ingress_flow,
                       egress_flow=egress_flow,
                       system=system,
                       hostname=hostname,
                       content_service=content_service,
                       collect=collect,
                       collected_dids=collected_dids,
                       memo=memo,
                       logger=logger)


class Content:
    """
    A Content class that holds information about a piece of content, including its name, segments, mediaType, and service.
    Attributes:
        name (str): The name of the content.
        segments (List<Segment): The list of segments in storage that make up the Content
        media_type (str): The media type of the content
        content_service (ContentService): A ContentService object used to retrieve the content data.
    """

    def __init__(self, name: str, segments: List[Segment], media_type: str, content_service: ContentService):
        self.name = name
        self.segments = segments
        self.media_type = media_type
        self.content_service = content_service

    def json(self):
        """
        Returns a dictionary representation of the Content object.

        Returns:
            dict: A dictionary containing 'name', 'segments', and 'mediaType' keys.
        """
        return {
            'name': self.name,
            'segments': [segment.json() for segment in self.segments],
            'mediaType': self.media_type
        }

    def copy(self):
        """
        Returns a deep copy of the Content object.

        Returns:
            Content: A deep copy of the Content object.
        """
        return Content(name=self.name,
                       segments=copy.deepcopy(self.segments),
                       media_type=self.media_type,
                       content_service=self.content_service)

    def subcontent(self, offset: int, size: int):
        """
        Returns a new Content object with a subset of the original content.

        Args:
            offset (int): The starting byte offset.
            size (int): The size of the subset in bytes.

        Returns:
            Content: A new Content object with the specified subcontent.
        """
        return Content(name=self.name,
                       segments=self.subsegments(offset, size),
                       media_type=self.media_type,
                       content_service=self.content_service)

    def subsegments(self, offset: int, size: int):
        if offset < 0:
            raise ValueError(f"subsegments offset must be positive, got {offset}")

        if size < 0:
            raise ValueError(f"subsegments size must be positive, got {size}")

        if size + offset > self.get_size():
            raise ValueError(f"Size + offset ({size} + {offset}) exceeds total Content size of {self.get_size()}")

        if size == 0:
            return []

        new_segments = []
        offset_remaining = offset
        size_remaining = size

        for segment in self.segments:
            if offset_remaining > 0:
                if segment.size < offset_remaining:
                    # the first offset is past this segment, skip it
                    offset_remaining -= segment.size
                    continue
                else:
                    # chop off the front of this segment
                    segment = Segment(uuid=segment.uuid,
                                      offset=segment.offset + offset_remaining,
                                      size=segment.size - offset_remaining,
                                      did=segment.did)
                    offset_remaining = 0

            if size_remaining < segment.size:
                # chop off the back of this segment
                segment = Segment(uuid=segment.uuid,
                                  offset=segment.offset,
                                  size=size_remaining,
                                  did=segment.did)
            size_remaining -= segment.size
            new_segments.append(segment)
            if size_remaining == 0:
                break

        return new_segments


    def get_size(self):
        """
        Returns the size of the content in bytes.

        Returns:
            int: The size of the content in bytes.
        """
        sum = 0
        for segment in self.segments:
            sum = sum + segment.size
        return sum

    def get_media_type(self):
        """
        Returns the media type of the content.

        Returns:
        str: The media type of the content.
        """
        return self.media_type

    def set_media_type(self, media_type: str):
        """
        Sets the media type of the content.

        Args:
            media_type (str): The media type to set.
        """
        self.media_type = media_type

    def load_bytes(self):
        """
        Retrieves the content as bytes.

        Returns:
            bytes: The content as bytes.
        """
        return self.content_service.get_bytes(self.segments)

    def load_str(self):
        """
        Retrieves the content as a string.

        Returns:
            str: The content as a string.
        """
        return self.content_service.get_str(self.segments)

    def prepend(self, other_content):
        """
        Prepends the content from another Content object.

        Args:
            other_content (Content): The Content object to prepend.
        """
        self.segments[0:0] = other_content.segments

    def append(self, other_content):
        """
        Appends the content from another Content object.

        Args:
            other_content (Content): The Content object to append.
        """
        self.segments.extend(other_content.segments)

    def __eq__(self, other):
        if isinstance(other, Content):
            return (self.name == other.name and
                    self.segments == other.segments and
                    self.media_type == other.media_type and
                    self.content_service == other.content_service)
        return False

    @classmethod
    def from_str(cls, context: Context, str_data: str, name: str, media_type: str):
        segment = context.content_service.put_str(context.did, str_data)
        return Content(name=name, segments=[segment], media_type=media_type, content_service=context.content_service)

    @classmethod
    def from_bytes(cls, context: Context, byte_data: bytes, name: str, media_type: str):
        segment = context.content_service.put_bytes(context.did, byte_data)
        return Content(name=name, segments=[segment], media_type=media_type, content_service=context.content_service)

    @classmethod
    def from_dict(cls, content: dict, content_service: ContentService):
        if 'name' in content:
            name = content['name']
        else:
            name = None
        segments = [Segment.from_dict(segment) for segment in content['segments']]
        media_type = content['mediaType']
        return Content(name=name,
                       segments=segments,
                       media_type=media_type,
                       content_service=content_service)


class Domain(NamedTuple):
    name: str
    value: str
    media_type: str

    @classmethod
    def from_dict(cls, domain: dict):
        name = domain['name']
        if 'value' in domain:
            value = domain['value']
        else:
            value = None
        media_type = domain['mediaType']
        return Domain(name=name,
                      value=value,
                      media_type=media_type)


class DeltaFileMessage(NamedTuple):
    metadata: Dict[str, str]
    content_list: List[Content]
    domains: List[Domain]
    enrichments: List[Domain]

    @classmethod
    def from_dict(cls, delta_file_message: dict, content_service: ContentService):
        metadata = delta_file_message['metadata']
        content_list = [Content.from_dict(content, content_service) for content in delta_file_message['contentList']]
        domains = [Domain.from_dict(domain) for domain in delta_file_message['domains']] if 'domains' in delta_file_message else []
        enrichments = [Domain.from_dict(domain) for domain in delta_file_message['enrichments']] if 'enrichments' in delta_file_message else []

        return DeltaFileMessage(metadata=metadata,
                                content_list=content_list,
                                domains=domains,
                                enrichments=enrichments)


class Event(NamedTuple):
    delta_file_messages: List[DeltaFileMessage]
    context: Context
    params: dict
    queue_name: str
    return_address: str

    @classmethod
    def create(cls, event: dict, hostname: str, content_service: ContentService, logger: Logger):
        delta_file_messages = [DeltaFileMessage.from_dict(delta_file_message, content_service) for delta_file_message in event['deltaFileMessages']]
        context = Context.create(event['actionContext'], hostname, content_service, logger)
        params = event['actionParams']
        queue_name = None
        if 'queueName' in event:
            queue_name = event['queueName']
        return_address = None
        if 'returnAddress' in event:
            return_address = event['returnAddress']
        return Event(delta_file_messages, context, params, queue_name, return_address)



#
#    DeltaFi - Data transformation and enrichment platform
#
#    Copyright 2021-2023 DeltaFi Contributors <deltafi@deltafi.org>
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

from typing import List

from deltafi.result import FormatResult, FormatManyResult

from .assertions import *
from .framework import TestCaseBase, ActionTest


class FormatTestCase(TestCaseBase):
    def __init__(self, fields: Dict):
        super().__init__(fields)
        self.metadata = {}
        self.delete_metadata_keys = []
        self.expected_format_many_result = []

    def expect_format_result(self, metadata: Dict, delete_metadata_keys: List[str]):
        self.expected_result_type = FormatResult
        self.metadata = metadata
        self.delete_metadata_keys = delete_metadata_keys

    def add_format_many_result(self, metadata: Dict, delete_metadata_keys: List):
        self.expected_result_type = FormatManyResult
        self.expected_format_many_result.append(
            {
                "metadata": metadata,
                "delete_metadata_keys": delete_metadata_keys
            }
        )


class FormatActionTest(ActionTest):
    def __init__(self, package_name: str):
        """
        Provides structure for testing DeltaFi Format action
        Args:
            package_name: name of the actions package for finding resources
        """
        super().__init__(package_name)

    def format(self, test_case: FormatTestCase):
        if test_case.expected_result_type == FormatManyResult:
            self.expect_format_many_result(test_case)
        elif test_case.expected_result_type == FormatResult:
            self.expect_format_result(test_case)
        else:
            super().execute(test_case)

    def expect_format_result(self, test_case: FormatTestCase):
        result = super().run_and_check_result_type(test_case, FormatResult)
        self.assert_format_result(test_case, result)

    def expect_format_many_result(self, test_case: FormatTestCase):
        result = super().run_and_check_result_type(test_case, FormatManyResult)
        self.assert_format_many_result(test_case, result)

    def assert_format_result(self, test_case: FormatTestCase, result: FormatResult):
        # Check output
        if result.content is None:
            self.compare_all_output(test_case.compare_tool, [])
        else:
            self.compare_all_output(test_case.compare_tool, [result.content])

        # Check metadata
        assert_keys_and_values(test_case.metadata, result.metadata)

        # Check deleted metadata
        for key in test_case.delete_metadata_keys:
            assert_key_in(key, result.delete_metadata_keys)

    def assert_format_many_result(self, test_case: FormatTestCase, actual: FormatManyResult):
        assert_equal_len(test_case.expected_format_many_result, actual.format_results)
        for index, expected_child_result in enumerate(test_case.expected_format_many_result):
            actual_child = actual.format_results[index]
            self.compare_one_content(test_case.compare_tool,
                                     self.expected_outputs[index],
                                     actual_child.format_result.content,
                                     index)

            assert_keys_and_values(expected_child_result['metadata'],
                                   actual_child.format_result.metadata)
            for key in expected_child_result['delete_metadata_keys']:
                assert_key_in(key, actual_child.format_result.delete_metadata_keys)



#
#    DeltaFi - Data transformation and enrichment platform
#
#    Copyright 2021-2023 DeltaFi Contributors <deltafi@deltafi.org>
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

import logging
import sys
from datetime import datetime

import json_logging


def get_logger(name: str = None) -> logging.Logger:
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)
    logger.addHandler(logging.StreamHandler(sys.stdout))
    logger.propagate = False

    if name is not None:
        logger = logging.LoggerAdapter(logger, dict(action=name))

    return logger


def _sanitize_log_msg(record):
    return record.getMessage().replace('\n', '_').replace('\r', '_').replace('\t', '_')


class JSONLogFormatter(json_logging.JSONLogFormatter):

    def _format_log_object(self, record, request_util):
        utcnow = datetime.utcnow()

        json_log_object = {
            'timestamp': json_logging.util.iso_time_format(utcnow),
            'message': _sanitize_log_msg(record),
            'loggerName': record.name,
            'threadName': record.threadName,
            'level': record.levelname,
            'module': record.module,
            'line_no': record.lineno
        }

        if hasattr(record, 'action'):
            json_log_object['action'] = record.action

        if hasattr(record, 'props'):
            json_log_object.update(record.props)

        if record.exc_info or record.exc_text:
            json_log_object.update(self.get_exc_fields(record))

        return json_log_object


json_logging.init_non_web(custom_formatter=JSONLogFormatter, enable_json=True)



#
#    DeltaFi - Data transformation and enrichment platform
#
#    Copyright 2021-2023 DeltaFi Contributors <deltafi@deltafi.org>
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

from abc import ABC, abstractmethod
from typing import Any, List

from deltafi.actiontype import ActionType
from deltafi.genericmodel import GenericModel
from deltafi.domain import Context, DeltaFileMessage
from deltafi.input import DomainInput, EgressInput, EnrichInput, FormatInput, LoadInput, TransformInput, ValidateInput
from deltafi.result import *
from pydantic import BaseModel


class Action(ABC):
    def __init__(self, action_type: ActionType, description: str, requires_domains: List[str],
                 requires_enrichments: List[str], valid_result_types: tuple):
        self.action_type = action_type
        self.description = description
        self.requires_domains = requires_domains
        self.requires_enrichments = requires_enrichments
        self.valid_result_types = valid_result_types
        self.action_execution = None

    @abstractmethod
    def build_input(self, context: Context, delta_file_message: DeltaFileMessage):
        pass

    def collect(self, action_inputs: List[Any]):
        raise RuntimeError(f"Collect is not supported for {self.__class__.__name__}")

    @abstractmethod
    def execute(self, context: Context, action_input: Any, params: BaseModel):
        pass

    def execute_action(self, event):
        if event.delta_file_messages is None or not len(event.delta_file_messages):
            raise RuntimeError(f"Received event with no delta file messages for did {event.context.did}")

        if event.context.collect is not None:
            result = self.execute(event.context, self.collect([self.build_input(event.context, delta_file_message)
                                                               for delta_file_message in event.delta_file_messages]),
                                  self.param_class().model_validate(event.params))
        else:
            result = self.execute(event.context, self.build_input(event.context, event.delta_file_messages[0]),
                                  self.param_class().model_validate(event.params))

        self.validate_type(result)
        return result

    @staticmethod
    def param_class( ):
        """Factory method to create and return an empty GenericModel instance.

        Returns
        -------
        GenericModel
            an empty GenericModel instance
        """
        return GenericModel

    def validate_type(self, result):
        if not isinstance(result, self.valid_result_types):
            raise ValueError(f"{self.__class__.__name__} must return one of "
                             f"{[result_type.__name__ for result_type in self.valid_result_types]} "
                             f"but a {result.__class__.__name__} was returned")


class DomainAction(Action, ABC):
    def __init__(self, description: str, requires_domains: List[str]):
        super().__init__(ActionType.DOMAIN, description, requires_domains, [], (DomainResult, ErrorResult))

    def build_input(self, context: Context, delta_file_message: DeltaFileMessage):
        return DomainInput(content=delta_file_message.content_list, metadata=delta_file_message.metadata,
                           domains={domain.name: domain for domain in delta_file_message.domains})

    @abstractmethod
    def domain(self, context: Context, params: BaseModel, domain_input: DomainInput):
        pass

    def execute(self, context: Context, domain_input: DomainInput, params: BaseModel):
        return self.domain(context, params, domain_input)


class EgressAction(Action, ABC):
    def __init__(self, description: str):
        super().__init__(ActionType.EGRESS, description, [], [], (EgressResult, ErrorResult, FilterResult))

    def build_input(self, context: Context, delta_file_message: DeltaFileMessage):
        return EgressInput(content=delta_file_message.content_list[0], metadata=delta_file_message.metadata)

    @abstractmethod
    def egress(self, context: Context, params: BaseModel, egress_input: EgressInput):
        pass

    def execute(self, context: Context, egress_input: EgressInput, params: BaseModel):
        return self.egress(context, params, egress_input)


class EnrichAction(Action, ABC):
    def __init__(self, description: str, requires_domains: List[str], requires_enrichments: List[str]):
        super().__init__(ActionType.ENRICH, description, requires_domains, requires_enrichments,
                         (EnrichResult, ErrorResult))

    def build_input(self, context: Context, delta_file_message: DeltaFileMessage):
        return EnrichInput(content=delta_file_message.content_list, metadata=delta_file_message.metadata,
                           domains={domain.name: domain for domain in delta_file_message.domains},
                           enrichments={domain.name: domain for domain in delta_file_message.enrichments})

    @abstractmethod
    def enrich(self, context: Context, params: BaseModel, enrich_input: EnrichInput):
        pass

    def execute(self, context: Context, enrich_input: EnrichInput, params: BaseModel):
        return self.enrich(context, params, enrich_input)


class FormatAction(Action, ABC):
    def __init__(self, description: str, requires_domains: List[str], requires_enrichments: List[str]):
        super().__init__(ActionType.FORMAT, description, requires_domains, requires_enrichments,
                         (FormatResult, FormatManyResult, ErrorResult, FilterResult))

    def build_input(self, context: Context, delta_file_message: DeltaFileMessage):
        return FormatInput(content=delta_file_message.content_list, metadata=delta_file_message.metadata,
                           domains={domain.name: domain for domain in delta_file_message.domains},
                           enrichments={domain.name: domain for domain in delta_file_message.enrichments})

    def collect(self, format_inputs: List[FormatInput]):
        all_content = []
        all_metadata = {}
        all_domains = {}
        all_enrichments = {}
        for format_input in format_inputs:
            all_content += format_input.content
            all_metadata.update(format_input.metadata)
            all_domains.update(format_input.domains)
            all_enrichments.update(format_input.enrichments)
        return FormatInput(content=all_content, metadata=all_metadata, domains=all_domains, enrichments=all_enrichments)

    @abstractmethod
    def format(self, context: Context, params: BaseModel, format_input: FormatInput):
        pass

    def execute(self, context: Context, format_input: FormatInput, params: BaseModel):
        return self.format(context, params, format_input)


class LoadAction(Action, ABC):
    def __init__(self, description: str):
        super().__init__(ActionType.LOAD, description, [], [],
                         (LoadResult, LoadManyResult, ErrorResult, FilterResult, ReinjectResult))

    def build_input(self, context: Context, delta_file_message: DeltaFileMessage):
        return LoadInput(content=delta_file_message.content_list, metadata=delta_file_message.metadata)

    def collect(self, load_inputs: List[LoadInput]):
        all_content = []
        all_metadata = {}
        for load_input in load_inputs:
            all_content += load_input.content
            all_metadata.update(load_input.metadata)
        return LoadInput(content=all_content, metadata=all_metadata)

    @abstractmethod
    def load(self, context: Context, params: BaseModel, load_input: LoadInput):
        pass

    def execute(self, context: Context, load_input: LoadInput, params: BaseModel):
        return self.load(context, params, load_input)


class TimedIngressAction(Action, ABC):
    def __init__(self, description: str):
        super().__init__(ActionType.TIMED_INGRESS, description, [], [], (IngressResult, ErrorResult))

    def build_input(self, context: Context, delta_file_message: DeltaFileMessage):
        return None

    @abstractmethod
    def ingress(self, context: Context, params: BaseModel):
        pass

    def execute(self, context: Context, input_placeholder: Any, params: BaseModel):
        return self.ingress(context, params)


class TransformAction(Action, ABC):
    def __init__(self, description: str):
        super().__init__(ActionType.TRANSFORM, description, [], [], (TransformResult, ErrorResult, FilterResult, ReinjectResult))

    def build_input(self, context: Context, delta_file_message: DeltaFileMessage):
        return TransformInput(content=delta_file_message.content_list, metadata=delta_file_message.metadata)

    def collect(self, transform_inputs: List[TransformInput]):
        all_content = []
        all_metadata = {}
        for transform_input in transform_inputs:
            all_content += transform_input.content
            all_metadata.update(transform_input.metadata)
        return TransformInput(content=all_content, metadata=all_metadata)

    @abstractmethod
    def transform(self, context: Context, params: BaseModel, transform_input: TransformInput):
        pass

    def execute(self, context: Context, transform_input: TransformInput, params: BaseModel):
        return self.transform(context, params, transform_input)


class ValidateAction(Action, ABC):
    def __init__(self, description: str):
        super().__init__(ActionType.VALIDATE, description, [], [], (ValidateResult, ErrorResult, FilterResult))

    def build_input(self, context: Context, delta_file_message: DeltaFileMessage):
        return ValidateInput(content=delta_file_message.content_list[0], metadata=delta_file_message.metadata)

    @abstractmethod
    def validate(self, context: Context, params: BaseModel, validate_input: ValidateInput):
        pass

    def execute(self, context: Context, validate_input: ValidateInput, params: BaseModel):
        return self.validate(context, params, validate_input)



#
#    DeltaFi - Data transformation and enrichment platform
#
#    Copyright 2021-2023 DeltaFi Contributors <deltafi@deltafi.org>
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

import abc
from enum import Enum
import uuid
from typing import Dict, List

from deltafi.domain import Content, Context
from deltafi.metric import Metric

ENDPOINT_TAG = "endpoint"
FILES_OUT = "files_out"
BYTES_OUT = "bytes_out"


class Result:
    __metaclass__ = abc.ABCMeta

    def __init__(self, result_key, result_type, context):
        self.result_key = result_key
        self.result_type = result_type
        self.metrics = []
        self.context = context

    @abc.abstractmethod
    def response(self):
        pass

    def add_metric(self, metric: Metric):
        self.metrics.append(metric)


class DomainResult(Result):
    def __init__(self, context: Context):
        super().__init__('domain', 'DOMAIN', context)
        self.annotations = {}

    def annotate(self, key: str, value: str):
        self.annotations[key] = value
        return self

    def response(self):
        return {
            'annotations': self.annotations
        }


class EgressResult(Result):
    def __init__(self, context: Context, destination: str, bytes_egressed: int):
        super().__init__(None, 'EGRESS', context)
        self.add_metric(Metric(FILES_OUT, 1, {ENDPOINT_TAG: destination}))
        self.add_metric(Metric(BYTES_OUT, bytes_egressed, {ENDPOINT_TAG: destination}))

    def response(self):
        return None


class EnrichResult(Result):
    def __init__(self, context: Context):
        super().__init__('enrich', 'ENRICH', context)
        self.enrichments = []
        self.annotations = {}

    def enrich(self, name: str, value: str, media_type: str):
        self.enrichments.append({
            'name': name,
            'value': value,
            'mediaType': media_type
        })
        return self

    def annotate(self, key: str, value: str):
        self.annotations[key] = value
        return self

    def response(self):
        return {
            'enrichments': self.enrichments,
            'annotations': self.annotations
        }


class ErrorResult(Result):
    def __init__(self, context: Context, error_cause: str, error_context: str):
        super().__init__('error', 'ERROR', context)
        self.error_cause = error_cause
        self.error_context = error_context
        self.annotations = {}

    def annotate(self, key: str, value: str):
        self.annotations[key] = value
        return self

    def response(self):
        return {
            'cause': self.error_cause,
            'context': self.error_context,
            'annotations': self.annotations
        }


class FilterResult(Result):
    def __init__(self, context: Context, filtered_cause: str, filtered_context: str=None):
        super().__init__('filter', 'FILTER', context)
        self.filtered_cause = filtered_cause
        self.filtered_context = filtered_context
        self.annotations = {}

    def annotate(self, key: str, value: str):
        self.annotations[key] = value
        return self

    def response(self):
        return {
            'message': self.filtered_cause,
            'context': self.filtered_context,
            'annotations': self.annotations
        }


class FormatResult(Result):
    def __init__(self, context: Context):
        super().__init__('format', 'FORMAT', context)
        self.content = None
        self.delete_metadata_keys = []
        self.metadata = {}

    def set_metadata(self, metadata: dict):
        self.metadata = metadata
        return self

    def add_metadata(self, key: str, value: str):
        self.metadata[key] = value
        return self

    def delete_metadata_key(self, key: str):
        self.delete_metadata_keys.append(key)
        return self

    def set_content(self, content: Content):
        self.content = content
        return self

    def save_string_content(self, string_data: str, name: str, media_type: str):
        segment = self.context.content_service.put_str(self.context.did, string_data)
        self.content = Content(name=name, segments=[segment], media_type=media_type,
                               content_service=self.context.content_service)
        return self

    def save_byte_content(self, byte_data: bytes, name: str, media_type: str):
        segment = self.context.content_service.put_bytes(self.context.did, byte_data)
        self.content = Content(name=name, segments=[segment], media_type=media_type,
                               content_service=self.context.content_service)
        return self

    def response(self):
        return {
            'content': self.content.json(),
            'metadata': self.metadata,
            'deleteMetadataKeys': self.delete_metadata_keys
        }


class ChildFormatResult:
    def __init__(self, format_result: FormatResult = None):
        self._did = str(uuid.uuid4())
        self.format_result = format_result

    @property
    def did(self):
        return self._did

    def response(self):
        res = self.format_result.response()
        res["did"] = self._did
        return res


class FormatManyResult(Result):
    def __init__(self, context: Context):
        super().__init__('formatMany', 'FORMAT_MANY', context)
        self.format_results = []

    def add_format_result(self, format_result):
        if isinstance(format_result, ChildFormatResult):
            self.format_results.append(format_result)
        else:
            self.format_results.append(ChildFormatResult(format_result))
        return self

    def response(self):
        return [format_result.response() for format_result in self.format_results]


class IngressResultItem:
    def __init__(self, context: Context, filename: str):
        self.context = context
        self.filename = filename
        self._did = str(uuid.uuid4())
        self.content = []
        self.metadata = {}

    @property
    def did(self):
        return self._did

    # content can be a single Content or a List[Content]
    def add_content(self, content):
        if content:
            if type(content) == list:
                self.content.extend(content)
            else:
                self.content.append(content)

        return self

    def save_string_content(self, string_data: str, name: str, media_type: str):
        segment = self.context.content_service.put_str(self.context.did, string_data)
        self.content.append(
            Content(name=name, segments=[segment], media_type=media_type, content_service=self.context.content_service))
        return self

    def save_byte_content(self, byte_data: bytes, name: str, media_type: str):
        segment = self.context.content_service.put_bytes(self.context.did, byte_data)
        self.content.append(
            Content(name=name, segments=[segment], media_type=media_type, content_service=self.context.content_service))
        return self

    def set_metadata(self, metadata: dict):
        self.metadata = metadata
        return self

    def add_metadata(self, key: str, value: str):
        self.metadata[key] = value
        return self

    def response(self):
        return {
            'did': self._did,
            'filename': self.filename,
            'metadata': self.metadata,
            'content': [content.json() for content in self.content]
        }


class IngressStatusEnum(Enum):
    HEALTHY = 'HEALTHY'
    DEGRADED = 'DEGRADED'
    UNHEALTHY = 'UNHEALTHY'


class IngressResult(Result):
    def __init__(self, context: Context):
        super().__init__('ingress', 'INGRESS', context)
        self.memo = None
        self.execute_immediate = False
        self.ingress_result_items = []
        self.status = IngressStatusEnum.HEALTHY
        self.statusMessage = None

    def add_item(self, ingress_result_item: IngressResultItem):
        self.ingress_result_items.append(ingress_result_item)
        return self

    def response(self):
        return {
            'memo': self.memo,
            'executeImmediate': self.execute_immediate,
            'ingressItems': [ingress_result_item.response() for ingress_result_item in self.ingress_result_items],
            'status': self.status.value,
            'statusMessage': self.statusMessage
        }


class LoadResult(Result):
    def __init__(self, context: Context):
        super().__init__('load', 'LOAD', context)
        self.content = []
        self.metadata = {}
        self.domains = []
        self.annotations = {}
        self.delete_metadata_keys = []

    # content can be a single Content or a List[Content]
    def add_content(self, content):
        if content:
            if type(content) == list:
                self.content.extend(content)
            else:
                self.content.append(content)

        return self

    def save_string_content(self, string_data: str, name: str, media_type: str):
        segment = self.context.content_service.put_str(self.context.did, string_data)
        self.content.append(
            Content(name=name, segments=[segment], media_type=media_type, content_service=self.context.content_service))
        return self

    def save_byte_content(self, byte_data: bytes, name: str, media_type: str):
        segment = self.context.content_service.put_bytes(self.context.did, byte_data)
        self.content.append(
            Content(name=name, segments=[segment], media_type=media_type, content_service=self.context.content_service))
        return self

    def set_metadata(self, metadata: dict):
        self.metadata = metadata
        return self

    def add_metadata(self, key: str, value: str):
        self.metadata[key] = value
        return self

    def add_domain(self, name: str, value: str, media_type: str):
        self.domains.append({
            'name': name,
            'value': value,
            'mediaType': media_type})
        return self

    def annotate(self, key: str, value: str):
        self.annotations[key] = value
        return self

    def delete_metadata_key(self, key: str):
        self.delete_metadata_keys.append(key)
        return self

    def response(self):
        return {
            'domains': self.domains,
            'content': [content.json() for content in self.content],
            'metadata': self.metadata,
            'annotations': self.annotations,
            'deleteMetadataKeys': self.delete_metadata_keys
        }


class ChildLoadResult:
    def __init__(self, load_result: LoadResult = None):
        self._did = str(uuid.uuid4())
        self.load_result = load_result

    @property
    def did(self):
        return self._did

    def response(self):
        res = self.load_result.response()
        res["did"] = self._did
        return res


class LoadManyResult(Result):
    def __init__(self, context: Context):
        super().__init__('loadMany', 'LOAD_MANY', context)
        self.load_results = []

    def add_load_result(self, load_result):
        if isinstance(load_result, ChildLoadResult):
            self.load_results.append(load_result)
        else:
            self.load_results.append(ChildLoadResult(load_result))
        return self

    def response(self):
        return [load_result.response() for load_result in self.load_results]


class ReinjectResult(Result):
    class ReinjectChild:
        def __init__(self, filename: str, flow: str, content: List[Content], metadata: Dict[str, str]):
            self.filename = filename
            self.flow = flow
            self.content = content
            self.metadata = metadata

        def json(self):
            return {
                'filename': self.filename,
                'flow': self.flow,
                'metadata': self.metadata,
                'content': [content.json() for content in self.content]
            }

    def __init__(self, context: Context):
        super().__init__('reinject', 'REINJECT', context)
        self.children = []

    def add_child(self, filename: str, flow: str, content: List[Content], metadata: Dict[str, str]):
        child = ReinjectResult.ReinjectChild(filename, flow, content, metadata)
        self.children.append(child)

    def response(self):
        return [child.json() for child in self.children]


class TransformResult(Result):
    def __init__(self, context: Context):
        super().__init__('transform', 'TRANSFORM', context)
        self.content = []
        self.metadata = {}
        self.annotations = {}
        self.delete_metadata_keys = []

    # content can be a single Content or a List[Content]
    def add_content(self, content):
        if content:
            if type(content) == list:
                self.content.extend(content)
            else:
                self.content.append(content)

        return self

    def save_string_content(self, string_data: str, name: str, media_type: str):
        segment = self.context.content_service.put_str(self.context.did, string_data)
        self.content.append(
            Content(name=name, segments=[segment], media_type=media_type, content_service=self.context.content_service))
        return self

    def save_byte_content(self, byte_data: bytes, name: str, media_type: str):
        segment = self.context.content_service.put_bytes(self.context.did, byte_data)
        self.content.append(
            Content(name=name, segments=[segment], media_type=media_type, content_service=self.context.content_service))
        return self

    def set_metadata(self, metadata: dict):
        self.metadata = metadata
        return self

    def add_metadata(self, key: str, value: str):
        self.metadata[key] = value
        return self

    def annotate(self, key: str, value: str):
        self.annotations[key] = value
        return self

    def delete_metadata_key(self, key: str):
        self.delete_metadata_keys.append(key)
        return self

    def response(self):
        return {
            'content': [content.json() for content in self.content],
            'metadata': self.metadata,
            'annotations': self.annotations,
            'deleteMetadataKeys': self.delete_metadata_keys
        }


class ValidateResult(Result):
    def __init__(self, context: Context):
        super().__init__(None, 'VALIDATE', context)

    def response(self):
        return None



#
#    DeltaFi - Data transformation and enrichment platform
#
#    Copyright 2021-2023 DeltaFi Contributors <deltafi@deltafi.org>
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#



"""GenericModel

Provides an empty subclass of pydantic.BaseModel.

Starting Pydantic v2, the BaseModel cannot be directly instantiated.  This class provides for instantiation of GenericModel objects that inherit from BaseModel.

This class does not define fields for validation or any other purpose.
"""


from pydantic import BaseModel



class GenericModel(BaseModel):
    pass
